import codecs
import csv
import logging
import os
import warnings
from abc import ABC
from datetime import datetime
from ftplib import FTP
from io import BytesIO
from io import StringIO
from operator import itemgetter
from typing import Iterable
from typing import List
from typing import Optional
from typing import Tuple
from typing import TypeVar
from typing import Union

import config
import paramiko
import pydantic
from more_itertools import one
from paramiko.client import SSHClient
from paramiko.sftp_client import SFTPClient
from pydantic import BaseModel
from pydantic import parse_obj_as
from ra_utils.apply import apply

T = TypeVar("T")

logger = logging.getLogger(__name__)


class FileSet(ABC):
    def __init__(self):
        super().__init__()
        self._settings = config.get_config()

    def get_import_filenames(self) -> List[str]:
        raise NotImplementedError("must be implemented by subclass")

    def get_modified_datetime(self, filename: str) -> datetime:
        raise NotImplementedError("must be implemented by subclass")

    def read_file(self, filename: str) -> List[str]:
        raise NotImplementedError("must be implemented by subclass")

    def write_file(self, filename: str, stream: StringIO, folder: Optional[str] = None):
        raise NotImplementedError("must be implemented by subclass")

    def _convert_stringio_to_bytesio(
        self, output: StringIO, encoding: str = "utf-8"
    ) -> BytesIO:
        """Convert StringIO object `output` to a BytesIO object using `encoding`"""
        output.seek(0)
        bytes_output = BytesIO()
        bytes_writer = codecs.getwriter(encoding)(bytes_output)
        bytes_writer.write(output.getvalue())
        bytes_output.seek(0)
        return bytes_output


class FTPFileSet(FileSet):
    def _get_ftp_connector(self) -> FTP:
        try:
            ftp = FTP(self._settings.ftp_url)
        except Exception as e:
            raise config.ImproperlyConfigured(
                "cannot connect to FTP server %r"
                % getattr(self._settings, "ftp_url", None)
            ) from e
        else:
            ftp.encoding = "utf-8"
            ftp.login(user=self._settings.ftp_user, passwd=self._settings.ftp_pass)
            ftp.cwd(self._settings.ftp_folder)
            return ftp

    def get_import_filenames(self) -> List[str]:
        ftp = self._get_ftp_connector()
        filenames = ftp.nlst()
        return filenames

    def get_modified_datetime(self, filename: str) -> datetime:
        """Read the 'modified' field from an FTP file"""
        ftp = self._get_ftp_connector()
        files = ftp.mlsd()
        found_file = one(filter(lambda x: x[0] == filename, files))
        filename, facts = found_file
        # String is on the form: "20210323153241.448"
        modify_string = facts["modify"][:-4]
        return datetime.strptime(modify_string, "%Y%m%d%H%M%S")

    def read_file(self, filename: str) -> List[str]:
        ftp = self._get_ftp_connector()
        lines: List[str] = []
        ftp.retrlines(f"RETR {filename}", lines.append)
        return lines

    def write_file(self, filename: str, stream: StringIO, folder: Optional[str] = None):
        ftp = self._get_ftp_connector()
        if folder:
            ftp.cwd(folder)
        result = ftp.storlines(
            f"STOR {filename}", self._convert_stringio_to_bytesio(stream)
        )
        ftp.close()
        return result


class SFTPFileSet(FileSet):
    def _get_sftp_connector(self) -> Tuple[SSHClient, SFTPClient]:
        try:
            ssh = paramiko.SSHClient()

            ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())

            pkey = paramiko.RSAKey.from_private_key_file(
                filename=self._settings.ftp_ssh_key_path,
                password=self._settings.ftp_ssh_key_pass,
            )
        except Exception as e:
            raise config.ImproperlyConfigured(
                "cannot connect to FTP server %r"
                % getattr(self._settings, "ftp_url", None)
            ) from e
        else:
            try:
                ssh.connect(
                    hostname=self._settings.ftp_url,
                    port=self._settings.ftp_port,
                    username=self._settings.ftp_user,
                    password=self._settings.ftp_pass,
                    pkey=pkey,
                )

                sftp = ssh.open_sftp()
                if not sftp:
                    raise

            except Exception as e:
                raise config.ImproperlyConfigured(
                    "cannot connect to FTP server %r"
                    % getattr(self._settings, "ftp_url", None)
                ) from e
            else:
                sftp.chdir(self._settings.ftp_folder)
                return ssh, sftp

    def get_import_filenames(self) -> List[str]:
        ssh, sftp = self._get_sftp_connector()
        filenames = sftp.listdir()
        sftp.close()
        ssh.close()
        return filenames

    def get_modified_datetime(self, filename: str) -> datetime:
        """Read the 'modified' field from an FTP file"""
        ssh, sftp = self._get_sftp_connector()
        file_stats = sftp.stat(filename)
        sftp.close()
        ssh.close()
        modified_time = file_stats.st_mtime
        if not modified_time:
            raise AttributeError
        return datetime.fromtimestamp(float(modified_time))

    def read_file(self, filename: str) -> List[str]:
        ssh, sftp = self._get_sftp_connector()
        file = sftp.file(filename=filename, mode="r")
        lines = file.readlines()
        file.close()
        sftp.close()
        ssh.close()
        return lines

    def write_file(self, filename: str, stream: StringIO, folder: Optional[str] = None):
        ssh, sftp = self._get_sftp_connector()
        if folder:
            try:
                sftp.chdir(folder)
            except IOError:
                sftp.mkdir(folder)
                sftp.chdir(folder)
        result = sftp.putfo(
            fl=self._convert_stringio_to_bytesio(stream), remotepath=filename
        )
        sftp.close()
        ssh.close()
        return result


class FSFileSet(FileSet):
    def _get_import_file_path(self, filename: str):
        return os.path.join(self._settings.import_csv_folder, filename)

    def _get_export_file_path(self, filename: str):
        return os.path.join(self._settings.queries_dir, filename)

    def get_import_filenames(self) -> List[str]:
        return os.listdir(self._settings.import_csv_folder)

    def get_modified_datetime(self, filename: str) -> datetime:
        path = self._get_import_file_path(filename)
        if os.path.exists(path):
            modified_time = datetime.fromtimestamp(os.path.getmtime(path))
            return modified_time
        raise ValueError("%r does not exist" % path)

    def read_file(self, filename: str) -> List[str]:
        path = self._get_import_file_path(filename)
        if os.path.exists(path):
            with open(path) as f:
                return f.readlines()
        return []

    def write_file(self, filename: str, stream: StringIO, folder: Optional[str] = None):
        if folder is not None:
            warnings.warn(
                "%r.write_file received unused `folder` argument (= %r)"
                % (self.__class__.__name__, folder)
            )
        path = self._get_export_file_path(filename)
        with open(path, "w") as f:
            stream.seek(0)
            return f.writelines(stream.getvalue())


def get_fileset_implementation() -> Union[SFTPFileSet, FTPFileSet, FSFileSet]:
    settings = config.get_config()
    if settings.import_csv_folder:
        return FSFileSet()
    elif settings.ftp_ssh_key_path:
        return SFTPFileSet()
    else:
        return FTPFileSet()


def parse_filenames(
    filenames: Iterable[str], prefix: str, last_import: datetime
) -> List[Tuple[str, datetime]]:
    """
    Get valid filenames matching a prefix and date newer than last_import

    All valid filenames are on the form: {{prefix}}_20210131_221600.csv
    """

    def parse_filepath(filepath: str) -> Tuple[str, datetime]:
        date_part = filepath[-19:-4]
        parsed_datetime = datetime.strptime(date_part, "%Y%m%d_%H%M%S")

        return filepath, parsed_datetime

    filtered_names = filter(lambda x: x.startswith(prefix), filenames)
    parsed_names = map(parse_filepath, filtered_names)
    # Only use files that are newer than last import, as agreed with Jesper Norup
    fileset = get_fileset_implementation()
    new_files = filter(
        apply(
            lambda filepath, filedate: fileset.get_modified_datetime(filepath)
            > last_import
        ),
        parsed_names,
    )
    sorted_filenames = sorted(new_files, key=itemgetter(1))
    return sorted_filenames


def parse_csv(lines: List[str], model: BaseModel) -> List[BaseModel]:
    def strip_empty(val: dict):
        return {k: v for k, v in val.items() if v != ""}

    reader = csv.DictReader(lines, delimiter=";")
    cleaned = list(map(strip_empty, reader))
    try:
        return parse_obj_as(List[model], cleaned)  # type: ignore
    except pydantic.ValidationError as e:
        raise ValueError("could not parse %r" % cleaned) from e


def read_csv(filename: str, model: T) -> List[T]:
    """Read CSV file from FTP into list of model objects"""
    logger.info("Processing %s", filename)
    fileset = get_fileset_implementation()
    lines = fileset.read_file(filename)
    return parse_csv(lines, model)  # type: ignore



# SPDX-FileCopyrightText: Magenta ApS
#
# SPDX-License-Identifier: MPL-2.0
"""Integration entrypoint."""
import logging
from contextlib import asynccontextmanager
from typing import Annotated
from typing import AsyncGenerator

import sentry_sdk
from fastapi import APIRouter
from fastapi import Depends
from fastapi import FastAPI
from fastramqpi.config import Settings as FastRAMQPISettings
from fastramqpi.depends import from_user_context
from fastramqpi.main import FastRAMQPI
from ramqp.depends import RateLimit
from ramqp.mo import MORouter
from ramqp.mo import MORoutingKey
from ramqp.mo import PayloadUUID

from .config import DatabaseSettings
from .config import GqlLoraCacheSettings
from .gql_lora_cache_async import GQLLoraCache
from .sql_export import SqlExport as _SqlExport
from .sql_table_defs import Adresse
from .sql_table_defs import Base
from .sql_table_defs import Bruger
from .sql_table_defs import DARAdresse
from .sql_table_defs import Engagement
from .sql_table_defs import Enhed
from .sql_table_defs import Enhedssammenkobling
from .sql_table_defs import Facet
from .sql_table_defs import ItForbindelse
from .sql_table_defs import ItSystem
from .sql_table_defs import Klasse
from .sql_table_defs import KLE
from .sql_table_defs import Leder
from .sql_table_defs import LederAnsvar
from .sql_table_defs import Orlov
from .sql_table_defs import Tilknytning
from .trigger import trigger_router

logger = logging.getLogger(__name__)

fastapi_router = APIRouter()
actualstate_router = MORouter()
historic_router = MORouter()

SqlExport = Annotated[_SqlExport, Depends(from_user_context("sql_exporter"))]
SqlExportHistoric = Annotated[
    _SqlExport, Depends(from_user_context("sql_exporter_historic"))
]


async def handle_address(uuid: PayloadUUID, sql_exporter: SqlExport):
    result = await sql_exporter.lc._fetch_address(uuid)
    address_objects = []
    dar_address_objects = []
    for res in result.get(str(uuid), []):
        if res["scope"] == "DAR":
            dar_address_objects.append(
                sql_exporter._generate_sql_dar_addresses(uuid, res, DARAdresse)
            )

        address_objects.append(sql_exporter._generate_sql_addresses(uuid, res, Adresse))

    sql_exporter.update_sql(uuid, dar_address_objects, DARAdresse)
    sql_exporter.update_sql(uuid, address_objects, Adresse)


async def handle_association(
    uuid: PayloadUUID,
    sql_exporter: SqlExport,
):
    result = await sql_exporter.lc._fetch_associations(uuid)

    association_objects = [
        sql_exporter._generate_sql_associations(uuid, res, Tilknytning)
        for res in result.get(str(uuid), [])
    ]

    sql_exporter.update_sql(uuid, association_objects, Tilknytning)


async def handle_class(
    uuid: PayloadUUID,
    sql_exporter: SqlExport,
):
    result = await sql_exporter.lc._fetch_classes(uuid)
    res = result.get(str(uuid))
    class_objects = (
        [sql_exporter._generate_sql_classes(uuid, res, Klasse)] if res else []
    )
    sql_exporter.update_sql(uuid, class_objects, Klasse)


async def handle_engagement(
    uuid: PayloadUUID,
    sql_exporter: SqlExport,
):
    result = await sql_exporter.lc._fetch_engagements(uuid)
    engagements_objects = [
        sql_exporter._generate_sql_engagements(uuid, res, Engagement)
        for res in result.get(str(uuid), [])
    ]

    sql_exporter.update_sql(uuid, engagements_objects, Engagement)


async def handle_facet(
    uuid: PayloadUUID,
    sql_exporter: SqlExport,
):
    result = await sql_exporter.lc._fetch_facets(uuid)
    res = result.get(str(uuid))
    facets_objects = (
        [sql_exporter._generate_sql_facets(uuid, res, Facet)] if res else []
    )

    sql_exporter.update_sql(uuid, facets_objects, Facet)


async def handle_it_system(
    uuid: PayloadUUID,
    sql_exporter: SqlExport,
):
    result = await sql_exporter.lc._fetch_itsystems(uuid)
    res = result.get(str(uuid))
    itsystems_objects = (
        [sql_exporter._generate_sql_it_systems(uuid, res, ItSystem)] if res else []
    )

    sql_exporter.update_sql(uuid, itsystems_objects, ItSystem)


async def handle_it_user(
    uuid: PayloadUUID,
    sql_exporter: SqlExport,
):
    result = await sql_exporter.lc._fetch_it_connections(uuid)
    it_connections_objects = [
        sql_exporter._generate_sql_it_user(uuid, res, ItForbindelse)
        for res in result.get(str(uuid), [])
    ]

    sql_exporter.update_sql(uuid, it_connections_objects, ItForbindelse)


async def handle_kle(
    uuid: PayloadUUID,
    sql_exporter: SqlExport,
):
    result = await sql_exporter.lc._fetch_kles(uuid)
    kle_objects = [
        sql_exporter._generate_sql_kle(uuid, res, KLE)
        for res in result.get(str(uuid), [])
    ]

    sql_exporter.update_sql(uuid, kle_objects, KLE)


async def handle_leave(
    uuid: PayloadUUID,
    sql_exporter: SqlExport,
):
    result = await sql_exporter.lc._fetch_leaves(uuid)
    leaves_objects = [
        sql_exporter._generate_sql_leave(uuid, res, Orlov)
        for res in result.get(str(uuid), [])
    ]

    sql_exporter.update_sql(uuid, leaves_objects, Orlov)


async def handle_manager(
    uuid: PayloadUUID,
    sql_exporter: SqlExport,
):
    result = await sql_exporter.lc._fetch_managers(uuid)
    managers_objects = []
    for res in result.get(str(uuid), []):
        managers_objects.append(sql_exporter._generate_sql_managers(uuid, res, Leder))

        for responsibility_uuid in res["manager_responsibility"]:
            manager_responsibility_objects = [
                sql_exporter._generate_sql_manager_responsibility(
                    responsibility_uuid, uuid, res, LederAnsvar
                )
                for res in result.get(str(uuid), [])
            ]
            sql_exporter.update_sql(uuid, manager_responsibility_objects, LederAnsvar)
    sql_exporter.update_sql(uuid, managers_objects, Leder)


async def handle_related(
    uuid: PayloadUUID,
    sql_exporter: SqlExport,
):
    result = await sql_exporter.lc._fetch_related(uuid)
    related_objects = [
        sql_exporter._generate_sql_related(uuid, res, Enhedssammenkobling)
        for res in result.get(str(uuid), [])
    ]

    sql_exporter.update_sql(uuid, related_objects, Enhedssammenkobling)


async def handle_org_unit(
    uuid: PayloadUUID,
    sql_exporter: SqlExport,
):
    result = await sql_exporter.lc._fetch_units(uuid)
    units_objects = [
        sql_exporter._generate_sql_units(uuid, res, Enhed)
        for res in result.get(str(uuid), [])
    ]

    sql_exporter.update_sql(uuid, units_objects, Enhed)


async def handle_person(
    uuid: PayloadUUID,
    sql_exporter: SqlExport,
):
    result = await sql_exporter.lc._fetch_users(uuid)
    users_objects = [
        sql_exporter._generate_sql_users(uuid, res, Bruger)
        for res in result.get(str(uuid), [])
    ]

    sql_exporter.update_sql(uuid, users_objects, Bruger)


handle_function_map = {
    "address": handle_address,
    "association": handle_association,
    "class": handle_class,
    "engagement": handle_engagement,
    "facet": handle_facet,
    "itsystem": handle_it_system,
    "ituser": handle_it_user,
    "kle": handle_kle,
    "leave": handle_leave,
    "manager": handle_manager,
    "related": handle_related,
    "org_unit": handle_org_unit,
    "person": handle_person,
}


@actualstate_router.register("address")
@actualstate_router.register("association")
@actualstate_router.register("class")
@actualstate_router.register("engagement")
@actualstate_router.register("facet")
@actualstate_router.register("itsystem")
@actualstate_router.register("ituser")
@actualstate_router.register("kle")
@actualstate_router.register("leave")
@actualstate_router.register("manager")
@actualstate_router.register("related")
@actualstate_router.register("org_unit")
@actualstate_router.register("person")
async def trigger_actual_state_event(
    uuid: PayloadUUID,
    sql_exporter: SqlExport,
    key: MORoutingKey,
    _: RateLimit,
):
    handle_function = handle_function_map[key]
    return await handle_function(uuid=uuid, sql_exporter=sql_exporter)


@historic_router.register("address")
@historic_router.register("association")
@historic_router.register("class")
@historic_router.register("engagement")
@historic_router.register("facet")
@historic_router.register("itsystem")
@historic_router.register("ituser")
@historic_router.register("kle")
@historic_router.register("leave")
@historic_router.register("manager")
@historic_router.register("related")
@historic_router.register("org_unit")
@historic_router.register("person")
async def trigger_historic_event(
    uuid: PayloadUUID,
    sql_exporter: SqlExportHistoric,
    key: MORoutingKey,
    _: RateLimit,
):
    handle_function = handle_function_map[key]
    return await handle_function(uuid=uuid, sql_exporter=sql_exporter)


class Settings(DatabaseSettings):
    fastramqpi: FastRAMQPISettings
    eventdriven: bool = False

    class Config:
        frozen = True
        env_nested_delimiter = "__"


@fastapi_router.get("/")
async def index() -> dict[str, str]:
    return {"name": "sql_export"}


def create_app(**kwargs) -> FastAPI:
    settings: Settings = Settings(**kwargs)
    settings.start_logging_based_on_settings()
    if settings.sentry_dsn:
        sentry_sdk.init(dsn=settings.sentry_dsn)

    fastramqpi = FastRAMQPI(application_name="sql-export", settings=settings.fastramqpi)
    if settings.eventdriven:
        amqpsystem = fastramqpi.get_amqpsystem()
        amqpsystem.router.registry.update(actualstate_router.registry)
        if settings.historic_state is not None:
            amqpsystem.router.registry.update(historic_router.registry)
    else:
        fastramqpi.get_app().include_router(trigger_router)
    fastramqpi.add_context(settings=settings, sql_exporter=None)

    app = fastramqpi.get_app()
    app.include_router(fastapi_router)

    context = fastramqpi.get_context()

    @asynccontextmanager
    async def sql_exporter(full_history) -> AsyncGenerator[None, None]:
        lc = GQLLoraCache(
            settings=GqlLoraCacheSettings().to_old_settings(),
            full_history=full_history,
            graphql_session=context["graphql_session"],
        )
        await lc._cache_lora_classes()
        await lc._cache_lora_facets()

        sql_exporter = SqlExport(
            settings=settings.to_old_settings(), historic=full_history
        )
        sql_exporter.lc = lc
        sql_exporter.session = sql_exporter._get_db_session()
        # Ensure that the tables exist
        # TODO: Once we only use event-driven sql-export we can delete the work-tables and "kvittering".
        # Then we can use create_all without the tables argument.
        Base.metadata.create_all(
            sql_exporter.engine,
            tables=[
                table
                for name, table in dict(Base.metadata.tables).items()
                if name[0] != "w" or name != "kvittering"
            ],
            checkfirst=True,
        )

        if full_history:
            fastramqpi.add_context(sql_exporter_historic=sql_exporter)
        else:
            fastramqpi.add_context(sql_exporter=sql_exporter)
        yield

    fastramqpi.add_lifespan_manager(sql_exporter(full_history=False), priority=1100)
    if settings.historic_state is not None:
        fastramqpi.add_lifespan_manager(sql_exporter(full_history=True), priority=1200)

    return fastramqpi.get_app()
